structure Parser :> PARSER =
  struct
    open Parsing
    open Tokens
    open Fomega

    infixr 4 << >>
    infixr 3 &&
    infix  2 -- ##
    infix  2 wth suchthat return guard when
    infixr 1 ||

    exception Internal of string
    exception Parse of Pos.pos * string
    fun error str = fn pos => raise Parse (pos, str)

    (* parse something and check that the predicate holds, with an error o/w *)
    infix 2 check

    (* val ` = literal *)
    fun `token = satisfy (fn x => eq (x, token))

    (* These are used to ensure that the identifier foo
       always maps to the same expvar foo_n *)
    structure ExpVarTable =
      SymTableFn(struct type symbol = string
                        type value = expvar
                        val compare = String.compare
                        val generate = Variable.newNamed
                 end)
    structure ConVarTable =
      SymTableFn(struct type symbol = string
                        type value = convar
                        val compare = String.compare
                        val generate = Variable.newNamed
                 end)
       
    val evTable = ExpVarTable.empty
    val cvTable = ConVarTable.empty

    (*
    fun ifmany f l = case l of [x] => x
                             | _ => f l

    (* get a marked thing's position *)
    fun pos (x, p) = p
    (* forget a marked thing's position *)
    fun forget (x, p) = x
    (* lift a binary function to return a marked thing *)
    fun mark f (xp, yp) = (f (xp, yp), Pos.union (pos xp, pos yp))
    (* lift a 2-item-list function to return a marked thing *)
    fun markl2 f l =
        case l of
            [xp, yp] => (f l, Pos.union (pos xp, pos yp))
          | _ => raise Internal "invalid argument to markl2"
    (* like ifmany, but ensures the result is unmarked *)
    fun ifmanymark f l = case l of [x] => forget x
                                 | _ => f l

    (* lifts Option.map to carry through a pos *)
    fun option_mapmark f opt =
        case opt of
            NONE => NONE
          | SOME (xp as (x, pos)) => SOME (f xp, pos)
    *)


    (*** Variables ***)

    val parseVar    = maybe (fn LCIDENT s => SOME (ExpVarTable.get evTable s)
                              | _ => NONE)
    val parseConVar = maybe (fn UCIDENT s => SOME (ConVarTable.get cvTable s)
                              | _ => NONE)


    (*** Kinds ***)

    fun parseAtomicKind () =
        alt [`TYPE return KndType,
             `LPAREN >>
                ($parseKind ## error "couldn't parse kind")
                << (`RPAREN ## error "expected `)'")]

    and parseArrowKind () =
        $parseAtomicKind
            && opt (`DARROW >> ($parseKind ## error "couldn't parse kind"))
            wth (fn (knd1, NONE) => knd1
                  | (knd1, SOME knd2) => KndArrow (knd1, knd2))

    and parseKind () = $parseArrowKind


    (*** Constructors ***)

    fun parseAtomicCon () =
        alt [parseConVar wth ConVar,
             `LPAREN >>
                ($parseCon ## error "couldn't parse con")
                << (`RPAREN ## error "expected `)'")]

    and parseAppCon () =
       (repeat1 ($parseAtomicCon)
        wth ListUtil.combinel (fn (cc, c) => ConApp (c, cc)))

    and parseArrowTy () =
        $parseAppCon
            && opt (`ARROW >> ($parseCon ## error "couldn't parse con"))
            wth (fn (ty1, NONE) => ty1
                  | (ty1, SOME ty2) => TyArrow (ty1, ty2))

    and parseCon () =
        alt [`ALL
                >> (parseConVar ## error "expected constructor variable")
                && (`COLONCOLON ## error "expected `::'")
                >> ($parseKind ## error "couldn't parse kind")
                && (`DOT ## error "expected `.'")
                >> ($parseCon ## error "couldn't parse con")
                wth TyForall o flat3,
             `LAMBDA
                >> (parseConVar ## error "expected constructor variable")
                && (`COLONCOLON ## error "expected `::'")
                >> ($parseKind ## error "couldn't parse kind")
                && (`DOT ## error "expected `.'")
                >> ($parseCon ## error "couldn't parse con")
                wth ConAbs o flat3,
             $parseArrowTy]
    (* XXX thought: Lambda vs lambda is confusing -- eliminate Lambda? *)

    
    (*** Expressions ***)

    fun parseAtomicExp () =
        alt [parseVar wth ExVar,
             `LPAREN >>
                ($parseExp ## error "couldn't parse exp")
                << (`RPAREN ## error "expected `)'")]

    and parseAppExp () =
        $parseAtomicExp
        && repeat ($parseAtomicExp wth Util.A
                || `LBRACKET >>
                        ($parseCon ## error "couldn't parse con")
                        << (`RBRACKET ## error "expected `]'")
                    wth Util.B)
        wth (fn (fexp, args) =>
                List.foldl (fn (Util.A e, f) => ExApp (f, e)
                             | (Util.B c, f) => ExConApp (f, c))
                           fexp
                           args)

    and parseExp () =
        alt [`LAMBDA
                >> (parseVar ## error "expected expression variable")
                && (`COLON ## error "expected `:'")
                >> ($parseCon ## error "couldn't parse con")
                && (`DOT ## error "expected `.'")
                >> ($parseExp ## error "couldn't parse exp")
                wth ExAbs o flat3,
             `BIGLAMBDA
                >> (parseConVar ## error "expected constructor variable")
                && (`COLONCOLON ## error "expected `::'")
                >> ($parseKind ## error "couldn't parse kind")
                && (`DOT ## error "expected `.'")
                >> ($parseExp ## error "couldn't parse exp")
                wth ExConAbs o flat3,
             $parseAppExp]



    (*** Programs (semicolon-delimited exps) ***)

    fun parseProgram () =
        separate' ($parseExp) (`SEMICOLON)
        << (done () ## error "expected eof")


    (*** for export ***)

    val parseKind = $parseKind
    val parseCon = $parseCon
    val parseExp = $parseExp
    val parseProgram = $parseProgram

    (*
    fun parse parser stream =
        Stream.hd (transform parser stream)
        handle Stream.Empty =>
            raise Parse (Pos.dummypos, "couldn't parse anything!")
    *)
    fun parse parser stream =
        case Parsing.parse parser stream of
            SOME x => x
          | NONE => raise Parse (Pos.dummypos, "couldn't parse anything!")

    fun parseString parser s =
        let val stream = StreamUtil.stostream s
            val mstream = Pos.markstream stream
            val tstream = Parsing.transform Tokenize.token mstream
        in
            parse parser tstream
        end

    fun parseFile parser file =
        let val stream = StreamUtil.ftostream file
            val mstream = Pos.markstreamex file stream
            val tstream = Parsing.transform Tokenize.token mstream
        in
            parse parser tstream
        end

    fun withHandler f x =
        f x
        handle
            Parse (pos, string) =>
                (print ("Syntax error: at " ^ Pos.toString pos ^ ": " ^ string);
                 print "\n";
                 raise Fail "done")
          | Tokenize.Lex (pos, string) =>
                (print ("Lexing error: at " ^ Pos.toString pos ^ ": " ^ string);
                 print "\n";
                 raise Fail "done")

  end
