
(* Combinator-style tokenizer. Takes characters as inputs, outputs
   items of type Tokens.token (see tokens.sml). This was adapted
   from a very old tokenizer for something else, so there are some
   aspects of it that are a little strange. *)

structure Tokenize :> TOKENIZE =
struct

  open Tokens
  open Parsing

  infixr 4 << >>
  infixr 3 &&
  infix  2 -- ##
  infix  2 wth suchthat return guard
  infixr 1 ||

  exception Lex of Pos.pos * string

  infix 2 check
  fun p check (pred, err) =
      !!p wth (fn (x, pos) => if pred x then x else raise Lex (pos, err x))

  (* Report an error with its position. *)
  fun error s pos = raise Lex (pos, s)

  (* Succeeds if p succeeds at the current point, but doesn't
     consume anything. *)

  (* Ignores the result of p *)
  fun ignore p = p return ()

  fun midchar #"_" = true
    | midchar #"'" = true
    | midchar _ = false

  val isSymbolic = Char.contains "!%&$#+-/:<=>@\\~`^|*"

  (* these never appear as part of a longer identifier *)
  val isSep = Char.contains "()[].; "

  fun identchar c = Char.isAlpha c orelse Char.isDigit c orelse midchar c

  exception Internal of string

  fun comment () = !!(string [#"(", #"*"]) -- (fn (_, p) => commentRest [p])

  and commentRest ps =
    alt [!!(string [#"(", #"*"]) -- (fn (_, p) => commentRest (p::ps)),
         string [#"*", #")"] -- (fn _ => case ps of
                                            [] => raise Internal "commentRest"
                                          | _::[] => succeed ()
                                          | _::ps => commentRest ps),
         any -- (fn _ => commentRest ps),
         done () -- (fn () => error "unterminated comment" (hd ps))]

  (* White space. *)

  val space = repeat (ignore ($comment) || ignore (satisfy Char.isSpace))

  val keywords =
      [("lambda", LAMBDA),
       ("Lambda", BIGLAMBDA),
       ("All", ALL),
       ("Type", TYPE),
       ("(", LPAREN),
       (")", RPAREN),
       ("[", LBRACKET),
       ("]", RBRACKET),
       (":", COLON),
       (";", SEMICOLON),
       ("::", COLONCOLON),
       ("=>", DARROW),
       ("->", ARROW),
       (".", DOT) ]

  fun ident (s, pos) =
      let
          fun id nil =
                let val c = String.sub (s, 0)
                in
                    if Char.isUpper c then
                        (UCIDENT s, pos)
                    else if Char.isLower c then
                        (LCIDENT s, pos)
                    else
                        raise Lex (pos, "illegal token: " ^ s)
                end
            | id ((a,b)::t) = if a = s then (b, pos) else id t
      in
          id keywords
      end

  val letters = satisfy Char.isAlpha && 
                repeat (satisfy identchar) wth op::

  val word = 
      alt [letters wth implode,
           (satisfy isSep) wth Char.toString,
           repeat1 (satisfy isSymbolic) wth implode,
           !!any wth (fn (c, pos) =>
                        raise Lex (pos, "illegal token: " ^ Char.toString c))]


  val token = space >> (!!word wth ident)
end
